# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


from .parser_manager import ParserManager


def parse(hex_version, bindings, include_dirs, is_strict=True):
    """ Parse a .sip file and return a 3-tuple of a Specification object, a
    list of Module objects and a list of the .sip files that specify the module
    to be generated.  A UserException is raised if there was an error.
    """

    return ParserManager(hex_version, bindings, include_dirs,
            is_strict).parse()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


cfg_package = 'imports'


def test_external_class(package):
    # Note that this only verifies that the generated code compiles but not
    # that it actually works.
    assert package.core_module.get_external() is None

def test_undefined_external_class(package):
    with pytest.raises(TypeError):
        package.core_module.check_undefined(None)

def test_superclasses(package):
    assert issubclass(package.extras_module.Subclass,
            package.core_module.BaseClass)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


CLASS_DOCSTRING = '''Klass overview

Klass()
Default constructor

Klass(other: Klass)
Copy constructor'''


OVERLOADED_DOCSTRINGS = '''overloaded()

overloaded(arg: bytes|None)
prepended

appended
overloaded(arg: int)

discarded'''


def test_module_docstring(module):
    assert module.__doc__ == 'Module'

def test_unbound_overloaded_docstrings(module):
    assert module.overloaded.__doc__ == OVERLOADED_DOCSTRINGS

def test_pod_ptr_arg_docstring(module):
    assert module.pod_ptr_arg.__doc__ == 'pod_ptr_arg(arg: bytes|None)'

def test_pod_ptr_opt_arg_docstring(module):
    assert module.pod_ptr_opt_arg.__doc__ == 'pod_ptr_opt_arg(arg: bytes|None = None)'

def test_bool_opt_arg_docstring(module):
    assert module.bool_opt_arg.__doc__ == 'bool_opt_arg(arg: bool = True)'

def test_class_docstring(module):
    assert module.Klass.__doc__ == CLASS_DOCSTRING

def test_class_instance_docstring(module):
    klass = module.Klass()

    assert klass.__doc__ == CLASS_DOCSTRING

def test_bound_class_callable(module):
    klass = module.Klass()

    assert klass.callable.__doc__ == 'callable'

def test_unbound_class_callable(module):
    assert module.Klass.callable.__doc__ == 'callable'



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


cfg_package = 'extenders'


def test_core_callable(package, abi_package_version):
    # Older versions of the ABI do not support extending existing callables
    # with new overloads.
    if abi_package_version >= 14:
        twice_i = package.core_module.NameSpace.twice(5)

        assert isinstance(twice_i, int)
        assert twice_i == 10

def test_core_callable_docstrings(package, abi_package_version):
    # Older versions of the ABI do not support dynamic docstrings.
    if abi_package_version >= 14:
        assert len(package.core_module.NameSpace.twice.__doc__.split('\n')) == 2

def test_extended_overload(package):
    twice_f = package.core_module.NameSpace.twice(5.0)

    assert isinstance(twice_f, float)
    assert twice_f == 10.0

def test_extended_callable_or_overload(package, abi_package_version):
    # Older versions of the ABI do not support extending existing callables
    # with new overloads.
    if abi_package_version >= 14:
        thrice_i = package.core_module.NameSpace.thrice(5)

        assert isinstance(thrice_i, int)
        assert thrice_i == 15

        thrice_f = package.core_module.NameSpace.thrice(5.0)

        assert isinstance(thrice_f, float)
        assert thrice_f == 15.0

def test_extended_callable_or_overload_docstrings(package, abi_package_version):
    # Older versions of the ABI do not support dynamic docstrings.
    if abi_package_version >= 14:
        assert len(package.core_module.NameSpace.thrice.__doc__.split('\n')) == 2

def test_extended_attribute_get(package):
    assert package.core_module.NameSpace.in_extras1 == 33

def test_extended_attribute_set(package, abi_package_version):
    package.core_module.NameSpace.in_extras1 = 44
    assert package.core_module.NameSpace.in_extras1 == 44

    # For ABI v14 check the C++ value has changed and not the type dict.
    if abi_package_version >= 14:
        assert package.extras1_module.get_in_extras1() == 44

def test_extended_operator_overload(package):
    klass = package.core_module.Klass()
    klass.set_int_attr(10)

    assert klass + 2.0 == 12.0

def test_extended_operator(package, abi_package_version):
    # Older versions of the ABI do not support new slots in extensions.
    if abi_package_version >= 14:
        klass = package.core_module.Klass()
        klass.set_int_attr(10)

        assert klass * 2 == 20

def test_init_extender(package):
    cast = package.extras1_module.Cast(20)
    klass = package.core_module.Klass(cast)

    assert klass.get_int_attr() == 20

def test_init_extender_docstrings(package, abi_package_version):
    # Older versions of the ABI do not support dynamic docstrings.
    if abi_package_version >= 14:
        assert len(package.core_module.Klass.__doc__.split('\n')) == 3

def test_dir(package):
    assert 'thrice' in dir(package.core_module.NameSpace)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


cfg_package = 'virtual_error_handlers'


def test_default_handler(package):
    class MyUseDefault(package.extras_module.UseDefault):
        def use_default(self):
            raise NotImplementedError

    inst = MyUseDefault()

    assert not hasattr(inst, 'default_used')
    inst.trigger_default()
    assert inst.default_used

def test_class_handler(package):
    class MyKlass(package.extras_module.Klass):
        def use_klass(self):
            raise NotImplementedError

    inst = MyKlass()

    assert not hasattr(inst, 'klass_used')
    inst.trigger_klass()
    assert inst.klass_used

def test_inst_handler(package):
    class MyKlass(package.extras_module.Klass):
        def use_inst(self):
            raise NotImplementedError

    inst = MyKlass()

    assert not hasattr(inst, 'inst_used')
    inst.trigger_inst()
    assert inst.inst_used

def test_no_handler(package, virtual_hook):
    class MyKlass(package.extras_module.Klass):
        def use_none(self):
            raise NotImplementedError

    inst = MyKlass()

    inst.trigger_none()
    with pytest.raises(NotImplementedError):
        virtual_hook.reraise()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


@pytest.fixture
def klass(module):
    """ This is a fixture that returns an instance of Klass. """

    return module.Klass()


@pytest.fixture
def reverse(module):
    """ This is a fixture that returns an instance of Reverse. """

    return module.Reverse()


def test_class_callables(module):
    assert module.Klass.get_s_attr_int() == 0
    module.Klass.set_s_attr_int(10)
    assert module.Klass.get_s_attr_int() == 10

def test_ctor_error(module):
    with pytest.raises(ValueError):
        module.Klass(10)

def test_method_error(klass):
    with pytest.raises(ValueError):
        klass.error_method()

def test_instance_callables(klass):
    assert klass.get_attr_int() == 0
    klass.set_attr_int(10)
    assert klass.get_attr_int() == 10

def test_slot_neg(klass):
    klass.set_attr_int(10)
    assert -klass == -10

def test_slot_sub(klass):
    klass.set_attr_int(10)
    assert klass - 5 == 5

def test_slot_add(klass):
    klass.set_attr_int(10)
    assert klass + 5 == 15

def test_slot_mul(klass):
    klass.set_attr_int(10)
    assert klass * 5 == 50

def test_slot_truediv(klass):
    klass.set_attr_int(10)
    assert klass / 3 == 3

def test_slot_mod(klass):
    klass.set_attr_int(10)
    assert klass % 3 == 1

def test_slot_and(klass):
    klass.set_attr_int(10)
    assert klass & 2 == 2

def test_slot_or(klass):
    klass.set_attr_int(10)
    assert klass | 5 == 15

def test_slot_xor(klass):
    klass.set_attr_int(10)
    assert klass ^ 1 == 11

def test_slot_invert(klass):
    klass.set_attr_int(10)
    assert ~klass == -11

def test_slot_lshift(klass):
    klass.set_attr_int(10)
    assert klass << 2 == 40

def test_slot_rshift(klass):
    klass.set_attr_int(10)
    assert klass >> 2 == 2

def test_slot_isub(klass):
    klass.set_attr_int(10)
    klass -= 5
    assert klass.get_attr_int() == 5

def test_slot_iadd(klass):
    klass.set_attr_int(10)
    klass += 5
    assert klass.get_attr_int() == 15

def test_slot_imul(klass):
    klass.set_attr_int(10)
    klass *= 5
    assert klass.get_attr_int() == 50

def test_slot_itruediv(klass):
    klass.set_attr_int(10)
    klass /= 3
    assert klass.get_attr_int() == 3

def test_slot_imod(klass):
    klass.set_attr_int(10)
    klass %= 3
    assert klass.get_attr_int() == 1

def test_slot_iand(klass):
    klass.set_attr_int(10)
    klass &= 2
    assert klass.get_attr_int() == 2

def test_slot_ior(klass):
    klass.set_attr_int(10)
    klass |= 5
    assert klass.get_attr_int() == 15

def test_slot_ixor(klass):
    klass.set_attr_int(10)
    klass ^= 1
    assert klass.get_attr_int() == 11

def test_slot_ilshift(klass):
    klass.set_attr_int(10)
    klass <<= 2
    assert klass.get_attr_int() == 40

def test_slot_irshift(klass):
    klass.set_attr_int(10)
    klass >>= 2
    assert klass.get_attr_int() == 2

def test_slot_eq(klass, module):
    klass.set_attr_int(10)
    assert klass == 10

def test_slot_ne(klass, module):
    klass.set_attr_int(10)
    assert klass != 1

def test_slot_lt(klass, module):
    klass.set_attr_int(10)
    assert klass < 100

def test_slot_le(klass, module):
    klass.set_attr_int(10)
    assert klass <= 10

def test_slot_gt(klass, module):
    klass.set_attr_int(10)
    assert klass > 1

def test_slot_ge(klass, module):
    klass.set_attr_int(10)
    assert klass >= 10

def test_slot_call(klass):
    klass.set_attr_int(33)
    assert klass(2) == 66

def test_slot_delitem(klass):
    original_count = klass.count()
    assert klass[2] == 2

    del klass[2]

    assert klass.count() == original_count - 1
    assert klass[2] == 3

def test_slot_getitem(klass):
    assert klass[2] == 2

    with pytest.raises(IndexError):
        klass[-1]

    with pytest.raises(IndexError):
        klass[klass.count()]

def test_slot_len(klass):
    assert klass.count() == len(klass)

def test_slot_setitem(klass):
    assert klass[2] == 2
    klass[2] = 20
    assert klass[2] == 20

    with pytest.raises(IndexError):
        klass[-1] = 0

    with pytest.raises(IndexError):
        klass[klass.count()] = 0

def test_slot_rsub(reverse):
    reverse.set_attr_int(10)
    assert 5 - reverse == -5

def test_slot_radd(reverse):
    reverse.set_attr_int(10)
    assert 5 + reverse == 15

def test_slot_rmul(reverse):
    reverse.set_attr_int(10)
    assert 5 * reverse == 50

def test_slot_rtruediv(reverse):
    reverse.set_attr_int(10)
    assert 16 / reverse == 1

def test_slot_rmod(reverse):
    reverse.set_attr_int(10)
    assert 16 % reverse == 6

def test_slot_rand(reverse):
    reverse.set_attr_int(10)
    assert 2 & reverse == 2

def test_slot_ror(reverse):
    reverse.set_attr_int(10)
    assert 5 | reverse == 15

def test_slot_rxor(reverse):
    reverse.set_attr_int(10)
    assert 1 ^ reverse == 11

def test_slot_rlshift(reverse):
    reverse.set_attr_int(2)
    assert 10 << reverse == 40

def test_slot_rrshift(reverse):
    reverse.set_attr_int(2)
    assert 10 >> reverse == 2



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


def test_ctor_with_transfer_arg(module):
    owned = module.OwnedClass()
    assert module.ispyowned(owned)

    klass = module.Klass(owned)
    assert not module.ispyowned(owned)

def test_static_callable_from_type(module):
    module.Klass.set_static(10)
    assert module.Klass.get_static() == 10

def test_static_callable_from_instance(module):
    klass = module.Klass()

    klass.set_static(20)
    assert klass.get_static() == 20

def test_bound_callable(module):
    klass = module.Klass()

    klass.set_inst(10)
    assert klass.get_inst() == 10

def test_unbound_callable(module):
    klass = module.Klass()

    module.Klass.set_inst(klass, 20)
    assert module.Klass.get_inst(klass) == 20

def test_protected_bound_callable(module):
    klass = module.Klass()

    klass.set_prot_inst(10)
    assert klass.get_prot_inst() == 10

def test_protected_unbound_callable(module):
    klass = module.Klass()

    module.Klass.set_prot_inst(klass, 20)
    assert module.Klass.get_prot_inst(klass) == 20



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


from ...specification import (AccessSpecifier, ArgumentType, CodeBlock,
        GILAction, IfaceFileType, MappedType, PyQtMethodSpecifier,
        WrappedClass)

from ..formatters import fmt_argument_as_cpp_type, fmt_class_as_scoped_name


def callable_overloads(member, overloads):
    """ A generator of the non-private and non-signal overloads. """

    for overload in overloads:
        if overload.common is member and overload.access_specifier is not AccessSpecifier.PRIVATE and overload.pyqt_method_specifier is not PyQtMethodSpecifier.SIGNAL:
            yield overload


def get_class_from_void(spec, klass):
    """ Return an assignment statement from a void * variable to a class
    instance variable.
    """

    klass_type = scoped_class_name(spec, klass)
    cast = get_type_from_void(spec, klass_type, 'sipCppV')

    return f'{klass_type} *sipCpp = {cast}'


def get_const_cast(spec, type, value):
    """ Return a value with an appropriate const_cast to a type. """

    if type.is_const:
        cpp_type = fmt_argument_as_cpp_type(spec, type, plain=True,
                no_derefs=True)

        return f'const_cast<{cpp_type} *>({value})'

    return value


def get_convert_to_type_code(type):
    """ Return a type's %ConvertToTypeCode. """

    if type.type in (ArgumentType.CLASS, ArgumentType.MAPPED) and not type.is_constrained:
        return type.definition.convert_to_type_code

    return None


def get_docstring_text(docstring):
    """ Return the text of a docstring. """

    text = docstring.text

    # Remove any single trailing newline.
    if text.endswith('\n'):
        text = text[:-1]

    s = ''

    for ch in text:
        if ch == '\n':
            # Let the compiler concatanate lines.
            s += '\\n"\n"'
        elif ch in r'\"':
            s += '\\'
            s += ch
        elif ch.isprintable():
            s += ch
        else:
            s += f'\\{ord(ch):03o}'

    return s


def get_encoded_type(module, klass, last=False):
    """ Return the structure representing an encoded type. """

    klass_module = klass.iface_file.module

    fields = [str(klass.iface_file.type_nr)]

    if klass_module is module:
        fields.append('255')
    else:
        for module_nr, imported_module in enumerate(module.all_imports):
            if imported_module is klass_module:
                fields.append(str(module_nr))
                break

    fields.append(str(int(last)))

    return '{' + ', '.join(fields) + '}'


def get_enum_class_scope(spec, enum):
    """ Return the scope of an unscoped enum as a string. """

    if enum.is_protected:
        scope_s = 'sip' + enum.scope.iface_file.fq_cpp_name.as_word
    else:
        scope_s = scoped_class_name(spec, enum.scope)

    return scope_s


def get_enum_member(spec, enum_member):
    """ Return an enum member as a string. """

    if spec.c_bindings:
        return enum_member.cpp_name

    enum = enum_member.scope

    if enum.no_scope:
        scope_s = ''
    else:
        if enum.is_scoped:
            scope_s = '::' + enum.cached_fq_cpp_name.name
        elif isinstance(enum.scope, WrappedClass):
            scope_s = get_enum_class_scope(spec, enum)
        elif isinstance(enum.scope, MappedType):
            scope_s = enum.scope.iface_file.fq_cpp_name.as_cpp
        else:
            # This can't happen.
            scope_s = ''

        scope_s += '::'

    return scope_s + enum_member.cpp_name


def get_function_table(members):
    """ Return a sorted list of relevant functions for a namespace. """

    return sorted(members, key=lambda m: m.py_name.name)


def get_mapped_type_flags(mapped_type):
    """ Return the flags for a mapped type. """

    flags = []

    if mapped_type.handles_none:
        flags.append('SIP_TYPE_ALLOW_NONE')

    if mapped_type.needs_user_state:
        flags.append('SIP_TYPE_USER_STATE')

    flags.append('SIP_TYPE_MAPPED')

    return '|'.join(flags)


def get_method_table(klass, ignore_slots=False):
    """ Return a sorted list of relevant methods (either lazy or non-lazy) for
    a class.
    """

    # Only provide an entry point if there is at least one overload that is
    # defined in this class and is a non-abstract function or slot.  We allow
    # private (even though we don't actually generate code) because we need to
    # intercept the name before it reaches a more public version further up the
    # class hierarchy.  We add the ctor and any variable handlers as special
    # entries.

    members = []

    for visible_member in klass.visible_members:
        if ignore_slots and visible_member.member.py_slot is not None:
            continue

        need_member = False

        for overload in visible_member.scope.overloads:
            # Skip protected methods if we don't have the means to handle them.
            if overload.access_specifier is AccessSpecifier.PROTECTED and not klass.has_shadow:
                continue

            if not skip_overload(overload, visible_member.member, klass, visible_member.scope):
                need_member = True

        if need_member:
            members.append(visible_member.member)

    return get_function_table(members)


def get_named_value_decl(spec, scope, type, name):
    """ Return the declaration of a named variable to hold a C++ value. """

    saved_derefs = type.derefs
    saved_is_const = type.is_const
    saved_is_reference = type.is_reference

    if len(type.derefs) == 0:
        if type.type in (ArgumentType.CLASS, ArgumentType.MAPPED):
            type.derefs = [False]
        else:
            type.is_const = False

    type.is_reference = False

    named_value_decl = fmt_argument_as_cpp_type(spec, type, name=name,
            scope=scope.iface_file if isinstance(scope, (WrappedClass, MappedType)) else None)

    type.derefs = saved_derefs
    type.is_const = saved_is_const
    type.is_reference = saved_is_reference

    return named_value_decl


def get_normalised_cached_name(cached_name):
    """ Return the normalised form of a cached name. """

    # If the name seems to be a template then just use the offset to ensure
    # that it is unique.
    if '<' in cached_name.name:
        return str(cached_name.offset)

    # Handle C++ and Python scopes.
    return cached_name.name.replace(':', '_').replace('.', '_')


def get_optional_ptr(is_ptr, name):
    """ Return an appropriate reference to an optional pointer. """

    return name if is_ptr else 'SIP_NULLPTR'


def get_type_from_void(spec, type_name, variable_name, derefs='*',
        tight=False):
    """ Return a cast from a void pointer variable to a pointer type. """

    if spec.c_bindings:
        if tight:
            return f'(({type_name} {derefs}){variable_name})'

        return f'({type_name} {derefs}){variable_name}'

    return f'reinterpret_cast<{type_name} {derefs}>({variable_name})'


def get_use_in_code(code, s, spec=None):
    """ Return the string to use depending on whether it is used in some code
    and optionally if the bindings are for C.
    """

    # Always use the string for C bindings.
    if spec is not None and spec.c_bindings:
        return s

    return s if is_used_in_code(code, s) else ''


def get_user_state_suffix(project, type):
    """ Return the suffix for functions that have a variant that supports a
    user state.
    """

    return 'US' if project.abi_version[0] >= 13 and type_needs_user_state(type) else ''


def get_void_ptr_cast(type):
    """ Return a void* cast for an argument if needed. """

    # Generate a cast if the argument's type was a typedef.  This allows us to
    # use typedef's to void* to hide something more complex that we don't
    # handle.
    if type.original_typedef is None:
        return ''

    return '(const void *)' if type.is_const else '(void *)'


def has_method_docstring(bindings, member, overloads):
    """ Return True if a function/method has a docstring. """

    auto_docstring = False

    # Check for any explicit docstrings and remember if there were any that
    # could be automatically generated.
    for overload in callable_overloads(member, overloads):
        if overload.docstring is not None:
            return True

        if bindings.docstrings:
            auto_docstring = True

    if member.no_arg_parser:
        return False

    return auto_docstring


def is_string(type):
    """ Check if a type is a string rather than a char type. """

    nr_derefs = len(type.derefs)

    if type.is_out and not type.is_reference:
        nr_derefs -= 1

    return nr_derefs > 0


def is_used_in_code(code, s):
    """ Return True if a string is used in code. """

    # The code may be a list of code blocks or an optional code block.
    if code is None:
        return False

    if isinstance(code, CodeBlock):
        code = [code]

    for cb in code:
        if s in cb.text:
            return True

    return False


# The types that need a Python reference.
_PY_REF_TYPES = (ArgumentType.ASCII_STRING, ArgumentType.LATIN1_STRING,
    ArgumentType.UTF8_STRING, ArgumentType.USTRING, ArgumentType.SSTRING,
    ArgumentType.STRING)

def keep_py_reference(project, arg):
    """ Return True if the argument has a type that requires an extra reference
    to the originating object to be kept.
    """

    if arg.is_reference or len(arg.derefs) == 0:
        return False

    if arg.type in _PY_REF_TYPES:
        return True

    # wchar_t strings/arrays don't leak in ABI v14 and later.  Note that
    # this solution could be adopted for earlier ABIs.
    return project.abi_version[0] >= 14 and arg.type is ArgumentType.WSTRING


def module_classes(spec):
    """ A generator of the classes defined in the module being generated. """

    module = spec.module

    for klass in spec.classes:
        if klass.iface_file.module is module:
            yield klass


def need_dealloc(spec, klass):
    """ Return True if a dealloc function is needed for a class. """

    if klass.iface_file.type is IfaceFileType.NAMESPACE:
        return False

    # Each of these conditions cause some code to be generated.

    if spec.bindings.tracing:
        return True

    if spec.c_bindings:
        return True

    if len(klass.dealloc_code) != 0:
        return True

    if klass.dtor is AccessSpecifier.PUBLIC:
        return True

    if klass.has_shadow:
        return True

    return False


def need_error_flag(code):
    """ Return True if handwritten code uses the error flag. """

    return is_used_in_code(code, 'sipError')


def py_scope(scope):
    """ Return the Python scope by accounting for hidden C++ namespaces. """

    return None if isinstance(scope, WrappedClass) and scope.is_hidden_namespace else scope


def pyqt5_supported(spec):
    """ Return True if the PyQt5 plugin was specified. """

    return 'PyQt5' in spec.plugins


def pyqt6_supported(spec):
    """ Return True if the PyQt6 plugin was specified. """

    return 'PyQt6' in spec.plugins


def release_gil(gil_action, bindings):
    """ Return True if the GIL is to be released. """

    return bindings.release_gil if gil_action is GILAction.DEFAULT else gil_action is GILAction.RELEASE


def scoped_class_name(spec, klass):
    """ Return a scoped class name as a string.  Protected classes have to be
    explicitly scoped.
    """

    return fmt_class_as_scoped_name(spec, klass, scope=klass.iface_file)


def scoped_variable_name(spec, variable):
    """ Return a scoped variable name as a string.  This should be used
    whenever the scope may be the instantiation of a template which specified
    /NoTypeName/.
    """

    scope = variable.scope
    fq_cpp_name = variable.fq_cpp_name

    if scope is None:
        return fq_cpp_name.as_cpp

    return scoped_class_name(spec, scope) + '::' + fq_cpp_name.base_name


def skip_overload(overload, member, klass, scope, want_local=True):
    """ See if a member overload should be skipped. """

    # Skip if it's not the right name.
    if overload.common is not member:
        return True

    # Skip if it's a signal.
    if overload.pyqt_method_specifier is PyQtMethodSpecifier.SIGNAL:
        return True

    # Skip if it's a private abstract.
    if overload.is_abstract and overload.access_specifier is AccessSpecifier.PRIVATE:
        return True

    # If we are disallowing them, skip if it's not in the current class unless
    # it is protected.
    if want_local and overload.access_specifier is not AccessSpecifier.PROTECTED and klass is not scope:
        return True

    return False


def type_needs_user_state(type):
    """ Return True if a type needs user state to be provided. """

    return type.type is ArgumentType.MAPPED and type.definition.needs_user_state


def variables_in_scope(spec, scope, check_handler=True):
    """ An iterator over the variables in a scope. """

    for variable in spec.variables:
        if py_scope(variable.scope) is scope and variable.module is spec.module:
            if check_handler and variable.needs_handler:
                continue

            yield variable



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


# Publish the API.  This is private to the rest of sip.
from .parser import parse
from .outputs import AbstractBackend
from .resolver import resolve



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from packaging.licenses import (canonicalize_license_expression,
        InvalidLicenseExpression)

from .exceptions import deprecated, UserFileException
from .py_versions import OLDEST_SUPPORTED_MINOR
from .toml import toml_loads


class PyProjectException(UserFileException):
    """ An exception related to a pyproject.toml file. """

    def __init__(self, text, *, detail=None):
        """ Initialise the exception. """

        super().__init__("pyproject.toml", text, detail=detail)


class PyProjectOptionException(PyProjectException):
    """ An exception related to a specific option of a pyproject.toml file. """

    def __init__(self, name, text, *, section_name=None, detail=None):
        """ Initialise the exception. """

        if section_name is None:
            section_name = 'tool.sip.project'

        super().__init__("'{0}.{1}': {2}".format(section_name, name, text),
                detail=detail)


class PyProjectTypeOptionException(PyProjectOptionException):
    """ An exception related to a badly typed option of a pyproject.toml file.
    """

    def __init__(self, name, types, *, section_name=None):
        """ Initialise the exception. """

        super().__init__(name, "value must be " + types,
                section_name=section_name)


class PyProjectUndefinedOptionException(PyProjectOptionException):
    """ An exception related to an undefined option of a pyproject.toml file.
    """

    def __init__(self, name, *, section_name=None):
        """ Initialise the exception. """

        super().__init__(name, "must be defined", section_name=section_name)


class PyProject:
    """ Encapsulate a parsed pyproject.toml file. """

    # The mapping of PEP 621 metadata to core metadata.
    _PEP621_CORE_MAP = {
        'name': 'name',
        'version': 'version',
        'description': 'summary',
        'requires-python': 'requires-python',
        'classifiers': 'classifier',
    }

    def __init__(self):
        """ Initialise the object. """

        # We remember any exception for now.
        try:
            with open('pyproject.toml', encoding='UTF-8') as f:
                self._raw_toml = f.read()

            self.pyproject = toml_loads(self._raw_toml)

        except FileNotFoundError:
            self._pending_exception = PyProjectException(
                    "there is no such file in the current directory")

        except UnicodeDecodeError as e:
            self._pending_exception = PyProjectException(
                    "is not a UTF-8 encoded file", details=str(e))

        except Exception as e:
            # Raise the exception now about a possibly badly formed file.
            self._pending_exception = PyProjectException(str(e))

        else:
            self._pending_exception = None

    def get_metadata(self):
        """ Return a dict containing the PEP 566 metadata. """

        # By this stage we have parsed the command line so stop if there has
        # been a problem.
        if self._pending_exception is not None:
            raise self._pending_exception

        # See if PEP 621 project metadata is supplied.  Make this required when
        # support for the legacy metadata is removed.
        metadata = self.get_section('project')
        if metadata is None:
            core_metadata = self._get_legacy_metadata()
        else:
            # The main validation we do is to check that the keys are defined
            # by PEP 621.
            core_metadata = {'metadata-version': '2.4'}

            for md_name, md_value in metadata.items():
                md_name = md_name.lower()

                # See if special handling is required.
                if md_name == 'readme':
                    self._handle_readme(md_value, core_metadata)
                elif md_name == 'license':
                    self._handle_license(md_value, core_metadata)
                elif md_name == 'license-files':
                    self._handle_license_files(md_value, core_metadata)
                elif md_name == 'authors':
                    self._handle_people(md_value, 'author', core_metadata)
                elif md_name == 'maintainers':
                    self._handle_people(md_value, 'maintainer', core_metadata)
                elif md_name == 'keywords':
                    self._handle_keywords(md_value, core_metadata)
                elif md_name == 'urls':
                    self._handle_urls(md_value, core_metadata)
                elif md_name == 'dependencies':
                    self._handle_dependencies(md_value, core_metadata)
                elif md_name == 'optional-dependencies':
                    self._handle_optional_dependencies(md_value, core_metadata)
                elif md_name in ('scripts', 'gui-scripts', 'entry-points'):
                    # We ignore these for the moment and stick with using
                    # [tool.sip.project].
                    pass
                else:
                    core_name = self._PEP621_CORE_MAP.get(md_name)
                    if core_name is None:
                        raise PyProjectOptionException(md_name,
                                "is an unsupported project key",
                                section_name='project')

                    core_metadata[core_name] = md_value

            # Check the legacy (ie. pre PEP 639) use of 'license'.
            if 'license' in core_metadata:
                if 'license-file' in core_metadata:
                    raise PyProjectOptionException('license',
                            "cannot be used in it's legacy form if 'license-files' is also specified",
                            section_name='project')

                deprecated(
                        "the legacy use of 'license'",
                        instead="an SPDX license expression and 'license-files'",
                        filename='pyproject.toml',
                        line_nr=self._get_line_nr('license ='))

                core_metadata['metadata-version'] = '2.2'

            # Check that the name has been specified.
            if 'name' not in core_metadata:
                raise PyProjectUndefinedOptionException('name',
                        section_name='project')

        if 'version' not in core_metadata:
            core_metadata['version'] = '0.1'

        if 'requires-python' not in core_metadata:
            # The minimal version of Python we support.
            core_metadata['requires-python'] = '>=3.{}'.format(
                    OLDEST_SUPPORTED_MINOR)

        return core_metadata

    def get_section(self, section_name, *, required=False):
        """ Return a sub-section with a dotted name. """

        # By this stage we may not have parsed the command line so pretend
        # everything is ok.
        if self._pending_exception is not None:
            return None

        section = self.pyproject

        for part in section_name.split('.'):
            try:
                section = section[part]
            except KeyError:
                if required:
                    raise PyProjectException(
                            "the '[{0}]' section is missing".format(
                                    section_name))

                return None

        if not self._is_section(section):
            raise PyProjectException(
                    "'{0}' is not a section".format(section_name))

        return section

    def _get_legacy_metadata(self):
        """ Return the pre-PEP 621 metadata. """

        line_nr = self._get_line_nr('[tool.sip.metadata]')

        if line_nr is None:
            # There is no metadata specified.
            raise PyProjectException("the '[project]' section is missing")

        deprecated(
                "using '[tool.sip.metadata]' to specify the project metadata",
                instead="'[project]'", filename='pyproject.toml',
                line_nr=line_nr)

        core_metadata = dict()
        name = None
        version = None
        core_metadata_version = None

        for md_name, md_value in self.get_section('tool.sip.metadata', required=True).items():
            md_name = md_name.lower()

            # Extract specific string values.
            if md_name in ('name', 'metadata-version'):
                if not isinstance(md_value, str):
                    raise PyProjectTypeOptionException(md_name, "a string",
                            section_name='tool.sip')

                if md_name == 'name':
                    if not md_value.replace('-', '_').isidentifier():
                        raise PyProjectOptionException('name',
                                "'{0}' is an invalid project name".format(
                                        md_value),
                                section_name='tool.sip')

                    name = md_value
                elif md_name == 'metadata-version':
                    core_metadata_version = md_value
            else:
                # Any other value may be a string or a list of strings.
                value_list = md_value if isinstance(md_value, list) else [md_value]

                for value in value_list:
                    if not isinstance(value, str):
                        raise PyProjectTypeOptionException(md_name,
                                "a string or a list of strings",
                                section_name='tool.sip')

            core_metadata[md_name] = md_value

        if name is None:
            raise PyProjectUndefinedOptionException('name',
                    section_name='tool.sip')

        if core_metadata_version is None:
            # Default to PEP 643.
            core_metadata['metadata-version'] = '2.2'

        return core_metadata

    def _get_line_nr(self, target):
        """ Return the number of the line in pyproject.toml containing a target
        string.
        """

        for line_nr, line in enumerate(self._raw_toml.split('\n')):
            line = line.strip()

            # Ignore comments.
            if line.startswith('#'):
                continue

            if target in line:
                return line_nr + 1

        # The target wasn't present.
        return None

    @classmethod
    def _handle_dependencies(cls, value, core_metadata):
        """ Handle the 'dependencies' key. """

        if not isinstance(value, list):
            raise PyProjectTypeOptionException('dependencies',
                    "an array of strings", section_name='project')

        cls._update_requires_dist(value, core_metadata)

    @staticmethod
    def _handle_keywords(value, core_metadata):
        """ Handle the 'keywords' key. """

        if not isinstance(value, list):
            raise PyProjectTypeOptionException('keywords',
                    "an array of strings", section_name='project')

        core_metadata['keywords'] = ', '.join(value)

    @classmethod
    def _handle_license(cls, value, core_metadata):
        """ Handle the 'license' key. """

        if isinstance(value, dict):
            return cls._handle_legacy_license(value, core_metadata)

        try:
            core_metadata['license-expression'] = canonicalize_license_expression(value)
        except InvalidLicenseExpression as e:
            raise PyProjectTypeOptionException('license',
                    "a valid SPDX license expression", section_name='project')

    @staticmethod
    def _handle_legacy_license(value, core_metadata):
        """ Handle the legacy 'license' key. """

        text = value.get('text')
        file = value.get('file')

        if text is not None:
            if file is not None:
                raise PyProjectOptionException('license',
                        "'file' and 'text' cannot both be specified",
                        section_name='project')
        elif file is not None:
            try:
                with open(file, encoding='UTF-8') as f:
                    text = f.read()
            except FileNotFoundError:
                raise PyProjectOptionException('file',
                        f"unable to read '{file}'",
                        section_name='project.license')
        else:
            raise PyProjectOptionException('readme',
                    "either 'file' or 'text' must be specified",
                    section_name='project')

        core_metadata['license'] = text

    @staticmethod
    def _handle_license_files(value, core_metadata):
        """ Handle the 'license-files' key. """

        if not isinstance(value, list):
            raise PyProjectTypeOptionException('license-files',
                    "an array of strings", section_name='project')

        core_metadata['license-file'] = value

    @classmethod
    def _handle_optional_dependencies(cls, value, core_metadata):
        """ Handle the 'optional-dependencies' key. """

        if not isinstance(value, dict):
            raise PyProjectTypeOptionException('optional-dependencies',
                    "a table", section_name='project')

        provides_extra = []
        requires_dist = []

        for extra, extra_deps in value.items():
            if not isinstance(extra_deps, list):
                raise PyProjectOptionException('optional-dependencies',
                        "each table value must be a list of strings",
                        section_name='project')

            provides_extra.append(extra)

            for dep in extra_deps:
                # Note that this is broken if the dependency is not a simple
                # name.
                requires_dist.append(f'{dep}[{extra}]')

        core_metadata['provides-extra'] = provides_extra
        cls._update_requires_dist(value, core_metadata)

    @staticmethod
    def _handle_people(value, role, core_metadata):
        """ Handle the people-related keys. """

        key = role + 's'

        if not isinstance(value, list):
            raise PyProjectTypeOptionException(key, "an array of tables",
                    section_name='project')

        all_names = []
        all_emails = []

        for table in value:
            if not isinstance(table, dict):
                raise PyProjectTypeOptionException(key, "an array of tables",
                        section_name='project')

            name = table.get('name')
            email = table.get('email')

            if name is None and email is None:
                raise PyProjectOptionException(key,
                        "each table must contain 'name' and/or 'email'",
                        section_name='project')

            if name is not None and not isinstance(name, str):
                raise PyProjectOptionException(key, "'name' must be a string",
                        section_name='project')

            if email is not None and not isinstance(email, str):
                raise PyProjectOptionException(key, "'email' must be a string",
                        section_name='project')

            if email is not None:
                if name is not None:
                    all_emails.append(f'{name} <{email}>')
                else:
                    all_emails.append(email)
            else:
                all_names.append(name)

        if all_names:
            core_metadata[role] = ', '.join(all_names)

        if all_emails:
            core_metadata[role + '-email'] = ', '.join(all_emails)

    @staticmethod
    def _handle_readme(value, core_metadata):
        """ Handle the 'readme' key. """

        if isinstance(value, str):
            readme = value.lower()

            if readme.endswith('.md'):
                content_type = 'text/markdown'
            elif readme.endswith('.rst'):
                content_type = 'text/x-rst'
            else:
                raise PyProjectOptionException('readme',
                        f"'{value}' has an unsupported file type",
                        section_name='project')

            description_file = value

        elif isinstance(value, dict):
            content_type = value.get('content-type')
            if content_type is None:
                raise PyProjectUndefinedOptionException('content-type',
                        section_name='project.readme')

            description_text = value.get('text')
            description_file = value.get('file')

            if description_text is not None:
                if description_file is not None:
                    raise PyProjectOptionException('readme',
                            "'file' and 'text' cannot both be specified",
                            section_name='project')

                core_metadata['description'] = text
            elif description_file is None:
                raise PyProjectOptionException('readme',
                        "either 'file' or 'text' must be specified",
                        section_name='project')

        else:
            raise PyProjectTypeOptionException('readme', "a string or a table",
                    section_name='project')

        core_metadata['description-content-type'] = content_type

        if description_file is not None:
            # This is a local, internal extension.
            core_metadata['description-file'] = description_file

    @staticmethod
    def _handle_urls(value, core_metadata):
        """ Handle the 'urls' key. """

        if not isinstance(value, dict):
            raise PyProjectTypeOptionException(key, "a table",
                    section_name='project')

        core_metadata['project-url'] = [f'{label}, {url}' for label, url in value.items()]

    @staticmethod
    def _is_section(value):
        """ Returns True if a section value is itself a section. """

        return isinstance(value, (dict, list))

    @staticmethod
    def _update_requires_dist(value, core_metadata):
        """ Update 'requires-dist' in the core metadata. """

        try:
            requires_dist = core_metadata['requires-dist']
        except KeyError:
            requires_dist = []

        requires_dist.extend(value)
        core_metadata['requires-dist'] = requires_dist



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest

from collections.abc import Buffer


def test_nonbuffer(module):
    nonbuffer = module.NonbufferClass()
    assert not isinstance(nonbuffer, Buffer)

def test_buffer(module):
    buffer = module.BufferClass(b'abc')
    mv = memoryview(buffer)
    assert mv.tobytes() == b'abc'



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


cfg_exceptions = True
cfg_package = 'imported_exceptions'


def test_imported_exceptions(package):
    with pytest.raises(package.handler_module.StdException):
        package.thrower_module.throwException()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


cfg_enabled_for = [13, 14]
cfg_package = 'v13_ns'


def test_wrapped_module_name(module):
    assert module.__name__ == 'v13_ns.v13_standalone_sip_module_module'

def test_toplevel_wrapped_types(module):
    assert module.Klass.__module__ == 'v13_ns.v13_standalone_sip_module_module'
    assert module.Klass.__name__ == 'Klass'
    assert module.Klass.__qualname__ == 'Klass'

def test_nested_wrapped_types(module):
    assert module.Klass.Nested.__module__ == 'v13_ns.v13_standalone_sip_module_module'
    assert module.Klass.Nested.__name__ == 'Nested'
    assert module.Klass.Nested.__qualname__ == 'Klass.Nested'

def test_sip_simplewrapper(module):
    from v13_ns.sip import simplewrapper, wrapper

    assert simplewrapper.__module__ == 'v13_ns.sip'
    assert simplewrapper.__name__ == 'simplewrapper'
    assert simplewrapper.__qualname__ == 'simplewrapper'

    assert issubclass(module.Klass.Nested, simplewrapper)
    assert not issubclass(module.Klass.Nested, wrapper)

def test_sip_wrapper(module):
    from v13_ns.sip import wrapper

    assert wrapper.__module__ == 'v13_ns.sip'
    assert wrapper.__name__ == 'wrapper'
    assert wrapper.__qualname__ == 'wrapper'

def test_sip_wrappertype(module):
    from v13_ns.sip import wrappertype

    assert wrappertype.__module__ == 'v13_ns.sip'
    assert wrappertype.__name__ == 'wrappertype'
    assert wrappertype.__qualname__ == 'wrappertype'



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


from sys import getrefcount

import pytest


def test_attrs_bool(module, abi_version):
    if abi_version >= 14:
        # Values are mapped to True and False.
        assert module.bool_attr is True
        module.bool_attr = False
        assert module.bool_attr is False
        module.bool_attr = 10
        assert module.bool_attr is True
        module.bool_attr = 0
        assert module.bool_attr is False
    else:
        # Values are left as integers.
        assert module.bool_attr
        module.bool_attr = False
        assert not module.bool_attr
        module.bool_attr = 10
        assert module.bool_attr
        module.bool_attr = 0
        assert not module.bool_attr

def test_attrs_byte(module):
    assert module.byte_attr == 10
    module.byte_attr = 20
    assert module.byte_attr == 20

def test_attrs_signed_byte(module):
    assert module.sbyte_attr == -10
    module.sbyte_attr = 20
    assert module.sbyte_attr == 20

def test_attrs_unsigned_byte(module):
    assert module.ubyte_attr == 10
    module.ubyte_attr = 20
    assert module.ubyte_attr == 20

def test_attrs_short(module):
    assert module.short_attr == -10
    module.short_attr = 20
    assert module.short_attr == 20

def test_attrs_unsigned_short(module):
    assert module.ushort_attr == 10
    module.ushort_attr = 20
    assert module.ushort_attr == 20

def test_attrs_int(module, abi_version):
    assert module.int_attr == -10
    module.int_attr = 20
    assert module.int_attr == 20

    # For ABI v14 check the C++ value has changed and not the module dict.
    if abi_version >= 14:
        assert module.get_int_attr() == 20

def test_attrs_unsigned_int(module):
    assert module.uint_attr == 10
    module.uint_attr = 20
    assert module.uint_attr == 20

def test_attrs_long(module):
    assert module.long_attr == -10
    module.long_attr = 20
    assert module.long_attr == 20

def test_attrs_unsigned_long(module):
    assert module.ulong_attr == 10
    module.ulong_attr = 20
    assert module.ulong_attr == 20

def test_attrs_long_long(module):
    assert module.longlong_attr == -10
    module.longlong_attr = 20
    assert module.longlong_attr == 20

def test_attrs_unsigned_long_long(module):
    assert module.ulonglong_attr == 10
    module.ulonglong_attr = 20
    assert module.ulonglong_attr == 20

def test_attrs_pyhasht(module):
    assert module.pyhasht_attr == -10
    module.pyhasht_attr = 20
    assert module.pyhasht_attr == 20

def test_attrs_pyssizet(module):
    assert module.pyssizet_attr == -10
    module.pyssizet_attr = 20
    assert module.pyssizet_attr == 20

def test_attrs_sizet(module):
    assert module.sizet_attr == 10
    module.sizet_attr = 20
    assert module.sizet_attr == 20

def test_attrs_float(module):
    assert module.float_attr == 10.
    module.float_attr = 20.
    assert module.float_attr == 20.

def test_attrs_double(module):
    assert module.double_attr == 10.
    module.double_attr = 20.
    assert module.double_attr == 20.

def test_attrs_char(module):
    assert module.char_attr == b'\x0a'
    module.char_attr = b'\x14'
    assert module.char_attr == b'\x14'

def test_attrs_char_ascii(module):
    assert module.char_ascii_attr == 'A'
    module.char_ascii_attr = 'Z'
    assert module.char_ascii_attr == 'Z'

def test_attrs_char_latin1(module):
    assert module.char_latin1_attr == '£'
    module.char_latin1_attr = '§'
    assert module.char_latin1_attr == '§'

def test_attrs_char_utf8(module):
    assert module.char_utf8_attr == 'A'
    module.char_utf8_attr = 'Z'
    assert module.char_utf8_attr == 'Z'

def test_attrs_signed_char(module):
    assert module.schar_attr == b'\x0a'
    module.schar_attr = b'\x14'
    assert module.schar_attr == b'\x14'

def test_attrs_unsigned_char(module):
    assert module.uchar_attr == b'\x0a'
    module.uchar_attr = b'\x14'
    assert module.uchar_attr == b'\x14'

def test_attrs_wchar(module):
    assert module.wchar_attr == 'β'
    module.wchar_attr = 'α'
    assert module.wchar_attr == 'α'

def test_attrs_string(module, abi_version):
    assert module.string_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.string_attr = b'bad'

def test_attrs_const_string(module):
    assert module.string_attr_const == b'str'
    module.string_attr_const = b'new_str'
    assert module.string_attr_const == b'new_str'
    module.string_attr_const = None
    assert module.string_attr_const is None

def test_attrs_string_ascii(module, abi_version):
    assert module.string_ascii_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.string_ascii_attr = 'bad'

def test_attrs_const_string_ascii(module, abi_version):
    assert module.string_ascii_attr_const == 'str'
    module.string_ascii_attr_const = 'new_str'
    assert module.string_ascii_attr_const == 'new_str'
    module.string_ascii_attr_const = None
    assert module.string_ascii_attr_const is None

    # ABI v14 will update the wrapped module attribute and convert the type in
    # the process.
    if abi_version >= 14:
        module.string_ascii_attr_const = b'bytes'
        assert module.string_ascii_attr_const == 'bytes'

def test_attrs_string_latin1(module, abi_version):
    assert module.string_latin1_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.string_latin1_attr = 'bad'

def test_attrs_const_string_latin1(module, abi_version):
    assert module.string_latin1_attr_const == '££'
    module.string_latin1_attr_const = '§§'
    assert module.string_latin1_attr_const == '§§'
    module.string_latin1_attr_const = None
    assert module.string_latin1_attr_const is None

    # ABI v14 will update the wrapped module attribute and convert the type in
    # the process.
    if abi_version >= 14:
        module.string_latin1_attr_const = '££'.encode('latin-1')
        assert module.string_latin1_attr_const == '££'

def test_attrs_string_utf8(module, abi_version):
    assert module.string_utf8_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.string_utf8_attr = 'bad'

def test_attrs_const_string_utf8(module, abi_version):
    assert module.string_utf8_attr_const == '2H₂ + O₂ ⇌ 2H₂O'
    module.string_utf8_attr_const = 'ሲተረጉሙ ይደረግሙ።'
    assert module.string_utf8_attr_const == 'ሲተረጉሙ ይደረግሙ።'
    module.string_utf8_attr_const = None
    assert module.string_utf8_attr_const is None

    # ABI v14 will update the wrapped module attribute and convert the type in
    # the process.
    if abi_version >= 14:
        module.string_utf8_attr_const = 'Καλημέρα'.encode('utf-8')
        assert module.string_utf8_attr_const == 'Καλημέρα'

def test_attrs_signed_string(module, abi_version):
    assert module.sstring_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.sstring_attr = b'bad'

def test_attrs_const_signed_string(module):
    assert module.sstring_attr_const == b'str'
    module.sstring_attr_const = b'new_str'
    assert module.sstring_attr_const == b'new_str'
    module.sstring_attr_const = None
    assert module.sstring_attr_const is None

def test_attrs_unsigned_string(module, abi_version):
    assert module.ustring_attr is None

    # For ABI v14 check a NULL wrapped module attribute cannot be updated to
    # point to potentially ephemeral data.  (Earlier versions just update the
    # module dictionary.)
    if abi_version >= 14:
        with pytest.raises(ValueError):
            module.ustring_attr = b'bad'

def test_attrs_const_unsigned_string(module):
    assert module.ustring_attr_const == b'str'
    module.ustring_attr_const = b'new_str'
    assert module.ustring_attr_const == b'new_str'
    module.ustring_attr_const = None
    assert module.ustring_attr_const is None

def test_attrs_voidptr(module):
    vp = module.voidptr_attr
    assert vp.asstring(size=5) == b'bytes'
    assert vp.getwriteable() is True

    module.voidptr_attr = None
    assert module.voidptr_attr is None

def test_attrs_const_voidptr(module, abi_version):
    const_vp = module.voidptr_const_attr
    assert const_vp.asstring(size=11) == b'bytes const'

    # ABIs prior to v14 don't honour the const of a wrapped module attribute.
    if abi_version >= 14:
        assert const_vp.getwriteable() is False

    module.voidptr_const_attr = None
    assert module.voidptr_const_attr is None

def test_attrs_pyobject(module):
    # Note that SIP does not check the Python types of these attributes (which
    # is really a bug) so we don't test the different types.
    assert module.pyobject_attr is None

    obj = object()
    obj_refcount = getrefcount(obj)

    module.pyobject_attr = obj
    assert module.pyobject_attr is obj
    assert getrefcount(obj) == obj_refcount + 1

def test_del_attrs(module, abi_version):
    # For ABI v14 check a wrapped module attribute cannot be deleted.
    if abi_version >= 14:
        with pytest.raises(AttributeError):
            del module.int_attr

def test_nonwrapped_attrs(module):
    with pytest.raises(AttributeError):
        module.foo

    module.foo = 'bar'
    assert module.foo == 'bar'

    del module.foo
    with pytest.raises(AttributeError):
        module.foo

def test_const_types(module, abi_version):
    # For ABI v14 check a wrapped const module attribute cannot be modified.
    if abi_version >= 14:
        assert module.int_attr_const == 10

        with pytest.raises(ValueError):
            module.int_attr_const = 0

def test_nonwrapped_attrs(module):
    module.nonwrapped_int = 10
    assert module.nonwrapped_int == 10

def test_py_name_annotation_attrs(module):
    assert module.py_int_attr == 10

def test_getters_and_setters(module, abi_version):
    if abi_version >= 14:
        assert module.int_attr_getter == 20

        module.int_attr_setter = 20
        assert module.int_attr_setter == 40

        with pytest.raises(NameError):
            module.int_attr_bad_setter = 0

        assert module.int_attr_no_setter == 10
        with pytest.raises(ValueError):
            module.int_attr_no_setter = 0

def test_dir(module):
    assert 'get_int_attr' in dir(module)

def test_dir_dynamic(module):
    assert 'bool_attr' in dir(module)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


from sys import getrefcount

import pytest


def test_enum_attrs(module, abi_version):
    """ Test a mapped typed containing an enum. """

    member = module.EnumAttrs.MTEnum.MTEnumMember
    value = member.value if abi_version >= 13 else member
    assert value == 10

def test_callables(module):
    """ Test a mapped typed containing a callable. """

    assert module.Callables.callable() == 42



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from abc import ABC, abstractmethod

from .configurable import Configurable


class AbstractBuilder(Configurable, ABC):
    """ This specifies the API of a builder. """

    def __init__(self, project, **kwargs):
        """ Initialise the builder. """

        super().__init__()

        self.project = project

        self.initialise_options(kwargs)

    @abstractmethod
    def build(self):
        """ Build the project in-situ. """

    @abstractmethod
    def build_sdist(self, sdist_directory):
        """ Build an sdist for the project and return the name of the sdist
        file.
        """

    @abstractmethod
    def build_wheel(self, wheel_directory):
        """ Build a wheel for the project and return the name of the wheel
        file.
        """

    @abstractmethod
    def install(self):
        """ Install the project. """



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>


import pytest


cfg_package = 'subclass_convertors'


def test_local_module(package):
    obj = package.core_module.Subclass1()
    obj.init_instance()

    assert isinstance(obj.get_instance(), package.core_module.Subclass1)

def test_imported_module(package):
    obj = package.extras_module.Subclass2()
    obj.init_instance()

    assert isinstance(obj.get_instance(), package.extras_module.Subclass2)

def test_needs_cast(package):
    obj = package.extras_module.Subclass4()
    obj.init_instance()

    assert isinstance(obj.get_instance(), package.extras_module.Subclass4)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2025 Phil Thompson <phil@riverbankcomputing.com>


import pytest


def test_instantiation(module):
    with pytest.raises(TypeError):
        module.NS()

def test_py_attributes(module):
    with pytest.raises(AttributeError):
        module.NS.foo

    module.NS.foo = 'bar'
    assert module.NS.foo == 'bar'

    del module.NS.foo
    with pytest.raises(AttributeError):
        module.NS.foo

def test_wrapped_attributes(module, abi_version):
    assert module.NS.attr == 0
    module.NS.attr = 10
    assert module.NS.attr == 10

    # For ABI v14 check the C++ value has changed and not the type dict.
    if abi_version >= 14:
        assert module.NS.get_attr() == 10

    # For ABI v14 check a wrapped namespace attribute cannot be deleted.
    if abi_version >= 14:
        with pytest.raises(AttributeError):
            del module.NS.attr



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2025 Phil Thompson <phil@riverbankcomputing.com>


from typing import Any, Generic, Iterable, overload, Sequence, TypeVar, Union


# PEP 484 has no explicit support for the buffer protocol so we just name types
# we know that implement it.
Buffer = Union[bytes, bytearray, memoryview, 'array', 'voidptr']


# Constants.
SIP_VERSION = ...       # type: int
SIP_VERSION_STR = ...   # type: str


# The bases for SIP generated types.
class wrappertype:
    def __init__(self, *args, **kwargs) -> None: ...

class simplewrapper:
    def __init__(self, *args, **kwargs) -> None: ...

class wrapper(simplewrapper): ...


# The array type.
_T = TypeVar('_T')

class array(Sequence[_T], Generic[_T]):

    @overload
    def __getitem__(self, key: int) -> _T: ...
    @overload
    def __getitem__(self, key: slice) -> 'array[_T]': ...

    @overload
    def __setitem__(self, key: int, value: _T) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Iterable[_T]) -> None: ...

    @overload
    def __delitem__(self, key: int) -> None: ...
    @overload
    def __delitem__(self, key: slice) -> None: ...

    def __len__(self) -> int: ...


# The voidptr type.
class voidptr:

    def __init__(self, addr: Union[int, Buffer], size: int = -1, writeable: bool = True) -> None: ...

    def __int__(self) -> int: ...

    @overload
    def __getitem__(self, i: int) -> bytes: ...

    @overload
    def __getitem__(self, s: slice) -> 'voidptr': ...

    def __len__(self) -> int: ...

    def __setitem__(self, i: Union[int, slice], v: Buffer) -> None: ...

    def asarray(self, size: int = -1) -> array[int]: ...

    # Python doesn't expose the capsule type.
    def ascapsule(self) -> Any: ...

    def asstring(self, size: int = -1) -> bytes: ...

    def getsize(self) -> int: ...

    def getwriteable(self) -> bool: ...

    def setsize(self, size: int) -> None: ...

    def setwriteable(self, writeable: bool) -> None: ...


# Remaining functions.
def assign(obj: simplewrapper, other: simplewrapper) -> None: ...
def cast(obj: simplewrapper, type: wrappertype) -> simplewrapper: ...
def delete(obj: simplewrapper) -> None: ...
def dump(obj: simplewrapper) -> None: ...
def enableautoconversion(type: wrappertype, enable: bool) -> bool: ...
def enableoverflowchecking(enable: bool) -> bool: ...
def getapi(name: str) -> int: ...
def isdeleted(obj: simplewrapper) -> bool: ...
def ispycreated(obj: simplewrapper) -> bool: ...
def ispyowned(obj: simplewrapper) -> bool: ...
def setapi(name: str, version: int) -> None: ...
def setdeleted(obj: simplewrapper) -> None: ...
def setdestroyonexit(destroy: bool) -> None: ...
def settracemask(mask: int) -> None: ...
def transferback(obj: wrapper) -> None: ...
def transferto(obj: wrapper, owner: wrapper) -> None: ...
def unwrapinstance(obj: simplewrapper) -> None: ...
def wrapinstance(addr: int, type: wrappertype) -> simplewrapper: ...



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2025 David Koňařík <dvdkon@konarici.cz>


from collections import defaultdict
from typing import TypeVar, TYPE_CHECKING

from .scoped_name import ScopedName

if TYPE_CHECKING:
    from .specification import (
        CachedName,
        MappedType,
        WrappedClass,
        WrappedEnum,
        WrappedTypedef,
    )


_T = TypeVar("_T")


class IndexedList(list[_T]):
    """
    A list of objects with methods overridden to maintain some indices of those
    objects. By itself, this class is just a list; subclasses can reimplement
    the underscored methods to actually create the indices.
    """

    # Abstract methods, to be implemented by subclass

    def _index_add(self, item):
        """Called when an item is added to the list."""
        ...

    def _index_remove(self, item):
        """Called when an item is removed from the list."""
        ...

    def _index_clear(self):
        """Called at initialisation and then whenever the list is cleared."""
        ...

    # Redefinitions of list methods to keep index up-to-date

    def __init__(self):
        """Create empty list."""
        super().__init__(self)
        self._index_clear()

    def append(self, v):
        """Add v to the end of list, updating indices in the process."""
        super().append(v)
        self._index_add(v)

    def insert(self, i, v):
        """Add v to the list at index i, updating indices in the process."""
        super().insert(i, v)
        self._index_add(v)

    def extend(self, vs):
        """
        Add elements of vs to the end of the list, updating indices in the
        process.
        """
        for v in vs:
            self.append(v)

    def remove(self, v):
        """
        Remove first occurrence of v from the list, updating indices in the
        process.
        """
        super().remove(v)
        self._index_remove(v)

    def clear(self):
        """
        Delete all elements from the list, updating indices in the process.
        """
        super().clear()
        self._index_clear()

    def __setitem__(self, i, v):
        """Set element at index i to v, updating indices in the process."""
        self._index_remove(self[i])
        super().__setitem__(i, v)
        self._index_add(v)

    def __delitem__(self, i):
        """Delete element at index i, updating indices in the process."""
        self._index_remove(self[i])
        super().__delitem__(i)


class IndexedClassList(IndexedList['WrappedClass']):
    """
    A list of WrappedClasss keeping the following indices:
    - classes by fq_cpp_name
    """

    def _index_clear(self):
        """Set up indices. See IndexedList._index_clear()."""

        self._by_cppname = defaultdict(list)

    def _index_add(self, klass):
        """Add element to indices. See IndexedList._index_add()."""

        self._by_cppname[klass.iface_file.fq_cpp_name].append(klass)

    def _index_remove(self, klass):
        """Remove element from indices. See IndexedList._index_add()."""

        self._by_cppname[klass.iface_file.fq_cpp_name].remove(klass)

    def by_fq_cpp_name(self, name):
        """Find WrappedClasses by their .iface_file.fq_cpp_name."""

        return self._by_cppname[name]


class IndexedEnumList(IndexedList['WrappedEnum']):
    """
    A list of WrappedEnums keeping the following indices:
    - enums by fq_cpp_name
    """

    def _index_clear(self):
        """ Initialise the indices. """

        self._by_cppname = {}

    def _index_add(self, enum):
        """ Add a WrappedEnum to the indices. """

        if enum.fq_cpp_name:
            self._by_cppname[enum.fq_cpp_name] = enum

    def _index_remove(self, enum):
        """ Remove a WrappedEnum from the indices. """

        if enum.fq_cpp_name:
            del self._by_cppname[enum.fq_cpp_name]

    def by_fq_cpp_name(self, name):
        """ Find a WrappedEnum by its .fq_cpp_name. """

        return self._by_cppname.get(name)


class IndexedCachedNameList(IndexedList['CachedName']):
    """
    A list of CachedNames keeping an index of the names by string values.
    """

    def _index_clear(self):
        """Set up indices. See IndexedList._index_clear()."""
        self._by_name = {}

    def _index_add(self, name):
        """Add CachedName to indices. See IndexedList._index_add()."""
        assert name.name not in self._by_name
        self._by_name[name.name] = name

    def _index_remove(self, name):
        """Remove CachedName from indices. See IndexedList._index_add()."""
        del self._by_name[name.name]

    def by_name(self, name: str):
        """Find CachedName by its .name."""
        return self._by_name.get(name)


class IndexedMappedTypeList(IndexedList['MappedType']):
    """
    A list of MappedTypes keeping an index of the names by readable base names.
    """

    def _index_clear(self):
        """Set up indices. See IndexedList._index_clear()."""
        self._by_readable_base_name = defaultdict(list)

    def _index_add(self, mapped_type):
        """Add MappedType to indices. See IndexedList._index_add()."""
        self._by_readable_base_name[
            mapped_type.iface_file.fq_cpp_name.readable_base_name
        ].append(mapped_type)

    def _index_remove(self, mapped_type):
        """Remove MappedType from indices. See IndexedList._index_add()."""
        self._by_readable_base_name[
            mapped_type.iface_file.fq_cpp_name.readable_base_name
        ].remove(mapped_type)

    def by_readable_base_name(self, name: str):
        """Find MappedTypes by their .iface_file.fq_cpp_name.readable_base_name."""
        return self._by_readable_base_name[name]


class IndexedTypedefList(IndexedList['WrappedTypedef']):
    """
    A list of WrappedTypedefs keeping an index of the typedefs by fq_cpp_name.
    """

    def _index_clear(self):
        """Set up indices. See IndexedList._index_clear()."""
        self._by_fq_cpp_name = {}

    def _index_add(self, typedef):
        """Add WrappedTypedef to indices. See IndexedList._index_add()."""
        name = tuple(typedef.fq_cpp_name._name)
        assert name not in self._by_fq_cpp_name
        self._by_fq_cpp_name[name] = typedef

    def _index_remove(self, typedef):
        """Remove WrappedTypedef from indices. See IndexedList._index_add()."""
        del self._by_fq_cpp_name[tuple(typedef.fq_cpp_name._name)]

    def by_fq_cpp_name(self, name: ScopedName):
        """Find WrappedTypedef by its .fq_cpp_name."""
        return self._by_fq_cpp_name.get(tuple(name._name))
